@@ -3,15 +3,6 @@
 '''
 A library to access Tryton's models like a client.
 '''
-import sys
-try:
-    import cdecimal
-    # Use cdecimal globally
-    if 'decimal' not in sys.modules:
-        sys.modules['decimal'] = cdecimal
-except ImportError:
-    import decimal
-    sys.modules['cdecimal'] = decimal
 import threading
 import datetime
 import functools
@@ -19,7 +10,7 @@ from decimal import Decimal
 
 import proteus.config
 
-__version__ = "4.6.4"
+__version__ = "5.0.0"
 __all__ = ['Model', 'Wizard', 'Report']
 
 _MODELS = threading.local()
@@ -56,7 +47,7 @@ class _EvalEnvironment(dict):
             pass
         return super(_EvalEnvironment, self).get(item, default)
 
-    def __nonzero__(self):
+    def __bool__(self):
         return True
 
     def __str__(self):
@@ -84,7 +75,7 @@ class dualmethod(object):
     >>> class Example(object):
     ...     @dualmethod
     ...     def method(cls, instances):
-    ...         print len(instances)
+    ...         print(len(instances))
     ...
     >>> Example.method([Example()])
     1
@@ -144,7 +135,7 @@ class CharDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert isinstance(value, basestring) or value is None
+        assert isinstance(value, str) or value is None
         super(CharDescriptor, self).__set__(instance, value or '')
 
 
@@ -159,14 +150,14 @@ class BinaryDescriptor(FieldDescriptor):
 class IntegerDescriptor(FieldDescriptor):
 
     def __set__(self, instance, value):
-        assert isinstance(value, (int, long, type(None)))
+        assert isinstance(value, (int, type(None)))
         super(IntegerDescriptor, self).__set__(instance, value)
 
 
 class FloatDescriptor(FieldDescriptor):
 
     def __set__(self, instance, value):
-        assert isinstance(value, (int, long, float, Decimal, type(None)))
+        assert isinstance(value, (int, float, Decimal, type(None)))
         if value is not None:
             value = float(value)
         super(FloatDescriptor, self).__set__(instance, value)
@@ -183,17 +174,20 @@ class NumericDescriptor(FieldDescriptor):
 class ReferenceDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         value = super(ReferenceDescriptor, self).__get__(instance, owner)
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             model_name, id = value.split(',', 1)
             if model_name:
-                relation = Model.get(model_name, instance._config)
-                value = relation(int(id))
+                Relation = Model.get(model_name, instance._config)
+                config = Relation._config
+                with config.reset_context(), \
+                        config.set_context(instance._context):
+                    value = Relation(int(id))
                 instance._values[self.name] = value
         return value
 
     def __set__(self, instance, value):
-        assert isinstance(value, (Model, type(None), basestring))
-        if isinstance(value, basestring):
+        assert isinstance(value, (Model, type(None), str))
+        if isinstance(value, str):
             assert value.startswith(',')
         elif isinstance(value, Model):
             assert value._config == instance._config
@@ -245,10 +239,12 @@ class DictDescriptor(FieldDescriptor):
 
 class Many2OneDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
-        relation = Model.get(self.definition['relation'], instance._config)
+        Relation = Model.get(self.definition['relation'], instance._config)
         value = super(Many2OneDescriptor, self).__get__(instance, owner)
-        if isinstance(value, (int, long)):
-            value = relation(value)
+        if isinstance(value, int):
+            config = Relation._config
+            with config.reset_context(), config.set_context(instance._context):
+                value = Relation(value)
         if self.name in instance._values:
             instance._values[self.name] = value
         return value
@@ -269,15 +265,16 @@ class One2ManyDescriptor(FieldDescriptor):
 
     def __get__(self, instance, owner):
         from .pyson import PYSONDecoder
-        relation = Model.get(self.definition['relation'], instance._config)
+        Relation = Model.get(self.definition['relation'], instance._config)
         value = super(One2ManyDescriptor, self).__get__(instance, owner)
         if not isinstance(value, ModelList):
             ctx = instance._context.copy() if instance._context else {}
             if self.definition.get('context'):
                 decoder = PYSONDecoder(_EvalEnvironment(instance))
                 ctx.update(decoder.decode(self.definition.get('context')))
-            with instance._config.set_context(ctx):
-                value = ModelList(self.definition, (relation(id)
+            config = Relation._config
+            with config.reset_context(), config.set_context(ctx):
+                value = ModelList(self.definition, (Relation(id)
                         for id in value or []), instance, self.name)
             instance._values[self.name] = value
         return value
@@ -459,7 +456,7 @@ class MetaModelFactory(object):
                 dict['_proxy'] = proxy
                 dict['_config'] = self.config
                 dict['_fields'] = proxy.fields_get(None, context)
-                for field_name, definition in dict['_fields'].iteritems():
+                for field_name, definition in dict['_fields'].items():
                     if field_name == 'id':
                         continue
                     Descriptor = self.descriptors.get(definition['type'],
@@ -498,13 +495,12 @@ class ModelList(list):
         self.domain = definition.get('domain', [])
         self.context = definition.get('context')
         self.add_remove = definition.get('add_remove')
+        self.search_order = definition.get('search_order', 'null')
+        self.search_context = definition.get('search_context', '{}')
         self.record_removed = set()
         self.record_deleted = set()
         result = super(ModelList, self).__init__(sequence)
-        for record in self:
-            record._parent = parent
-            record._parent_field_name = parent_field_name
-            record._parent_name = self.parent_name
+        self.__check(self, on_change=False)
         return result
     __init__.__doc__ = list.__init__.__doc__
 
@@ -521,10 +517,11 @@ class ModelList(list):
         ctx.update(decoder.decode(self.context) if self.context else {})
         return ctx
 
-    def __check(self, records):
+    def __check(self, records, on_change=True):
         config = None
         for record in records:
             assert isinstance(record, Model)
+            assert record.__class__.__name__ == self.model_name
             if self.parent:
                 assert record._config == self.parent._config
             elif self:
@@ -533,16 +530,14 @@ class ModelList(list):
                 assert record._config == config
             else:
                 config = record._config
+            if record._group is not self:
+                assert record._group is None
+                record._group = self
         for record in records:
-            assert record._parent is None
-            assert not record._parent_field_name
-            assert not record._parent_name
-            record._parent = self.parent
-            record._parent_field_name = self.parent_field_name
-            record._parent_name = self.parent_name
-
             # Set parent field to trigger on_change
-            if self.parent and self.parent_name in record._fields:
+            if (on_change
+                    and self.parent
+                    and self.parent_name in record._fields):
                 definition = record._fields[self.parent_name]
                 if definition['type'] in ('many2one', 'reference'):
                     setattr(record, self.parent_name, self.parent)
@@ -570,9 +565,7 @@ class ModelList(list):
 
     def pop(self, index=-1):
         self.record_removed.add(self[index])
-        self[index]._parent = None
-        self[index]._parent_field_name = ''
-        self[index]._parent_name = ''
+        self[index]._group = None
         res = super(ModelList, self).pop(index)
         self._changed()
         return res
@@ -581,9 +574,7 @@ class ModelList(list):
     def remove(self, record, _changed=True):
         if record.id >= 0:
             self.record_deleted.add(record)
-        record._parent = None
-        record._parent_field_name = ''
-        record._parent_name = ''
+        record._group = None
         res = super(ModelList, self).remove(record)
         if _changed:
             self._changed()
@@ -601,8 +592,10 @@ class ModelList(list):
     def new(self, **kwargs):
         'Adds a new record to the ModelList and returns it'
         Relation = Model.get(self.model_name, self.parent._config)
-        with Relation._config.set_context(self._get_context()):
-            new_record = Relation(**kwargs)
+        config = Relation._config
+        with config.reset_context(), config.set_context(self._get_context()):
+            # Set parent for on_change calls from default_get
+            new_record = Relation(_group=self, **kwargs)
         self.append(new_record)
         return new_record
 
@@ -617,7 +610,11 @@ class ModelList(list):
         add_remove_domain = (decoder.decode(self.add_remove)
             if self.add_remove else [])
         new_domain = [field_domain, add_remove_domain, condition]
-        with Relation._config.set_context(self._get_context()):
+        context = self._get_context()
+        context.update(decoder.decode(self.search_context))
+        order = order if order else decoder.decode(self.search_order)
+        config = Relation._config
+        with config.reset_context(), config.set_context(context):
             return Relation.find(new_domain, offset, limit, order)
 
     def set_sequence(self, field='sequence'):
@@ -659,7 +656,7 @@ class Model(object):
     _config = None
     _fields = None
 
-    def __init__(self, id=None, _default=True, **kwargs):
+    def __init__(self, id=None, _default=True, _group=None, **kwargs):
         super(Model, self).__init__()
         if id:
             assert not kwargs
@@ -668,20 +665,18 @@ class Model(object):
             Model.__counter -= 1
         self._values = {}  # store the values of fields
         self._changed = set()  # store the changed fields
-        self._parent = None  # store the parent record
-        self._parent_field_name = ''  # store the field name in parent record
-        self._parent_name = ''  # store the field name to parent record
-        self._context = self._config.context  # store the context
+        self._group = _group  # store the parent group
+        self.__context = self._config.context  # store the context
         if self.id < 0 and _default:
             self._default_get()
 
-        for field_name, value in kwargs.iteritems():
+        for field_name, value in kwargs.items():
             definition = self._fields[field_name]
             if definition['type'] in ('one2many', 'many2many'):
-                relation = Model.get(definition['relation'])
+                relation = Model.get(definition['relation'], self._config)
 
                 def instantiate(v):
-                    if isinstance(v, (int, long)):
+                    if isinstance(v, int):
                         return relation(v)
                     elif isinstance(v, dict):
                         return relation(_default=_default, **v)
@@ -691,20 +686,46 @@ class Model(object):
                 getattr(self, field_name).extend(value)
             else:
                 if definition['type'] == 'many2one':
-                    if isinstance(value, (int, long)):
-                        relation = Model.get(definition['relation'])
+                    if isinstance(value, int):
+                        relation = Model.get(
+                            definition['relation'], self._config)
                         value = relation(value)
                 setattr(self, field_name, value)
     __init__.__doc__ = object.__init__.__doc__
 
+    @property
+    def _parent(self):
+        if self._group is not None:
+            return self._group.parent
+
+    @property
+    def _parent_field_name(self):
+        if self._group is not None:
+            return self._group.parent_field_name
+        return ''
+
+    @property
+    def _parent_name(self):
+        if self._group is not None:
+            return self._group.parent_name
+        return ''
+
+    @property
+    def _context(self):
+        if self._group:
+            context = self._group._get_context()
+        else:
+            context = self.__context
+        return context
+
     @classmethod
     def get(cls, name, config=None):
         'Get a class for the named Model'
-        if (bytes == str) and isinstance(name, unicode):
+        if (bytes == str) and isinstance(name, str):
             name = name.encode('utf-8')
 
-        class Spam(Model):
-            __metaclass__ = MetaModelFactory(name, config=config)()
+        class Spam(Model, metaclass=MetaModelFactory(name, config=config)()):
+            pass
         return Spam
 
     @classmethod
@@ -762,7 +783,7 @@ class Model(object):
         if condition is None:
             condition = []
         ids = cls._proxy.search(condition, offset, limit, order,
-                cls._config.context)
+            cls._config.context)
         return [cls(id) for id in ids]
 
     @dualmethod
@@ -779,11 +800,12 @@ class Model(object):
             return
         proxy = records[0]._proxy
         config = records[0]._config
-        context = config.context
+        context = records[0]._context.copy()
         create, write = [], []
         for record in records:
             assert proxy == record._proxy
             assert config == record._config
+            assert context == record._context
             if record.id < 0:
                 create.append(record)
             elif record._changed:
@@ -813,15 +835,17 @@ class Model(object):
             return
         proxy = records[0]._proxy
         config = records[0]._config
-        context = config.context
-        context['_timestamp'] = {}
+        context = records[0]._context.copy()
+        timestamp = {}
         delete = []
         for record in records:
             assert proxy == record._proxy
             assert config == record._config
+            assert context == record._context
             if record.id > 0:
-                context['_timestamp'].update(record._get_timestamp())
+                timestamp.update(record._get_timestamp())
                 delete.append(record.id)
+        context['_timestamp'] = timestamp
         if delete:
             proxy.delete(delete, context)
         cls.reload(records)
@@ -840,7 +864,13 @@ class Model(object):
             return
 
         proxy = records[0]._proxy
-        context = records[0]._config.context
+        config = records[0]._config
+        context = records[0]._context.copy()
+        for record in records:
+            assert proxy == record._proxy
+            assert config == record._config
+            assert context == record._context
+
         if change is None:
             cls.save(records)
             cls.reload(records)  # Force reload because save doesn't always
@@ -867,7 +897,7 @@ class Model(object):
         'Return dictionary with timestamps'
         result = {'%s,%s' % (self.__class__.__name__, self.id):
                 self._timestamp}
-        for field, definition in self._fields.iteritems():
+        for field, definition in self._fields.items():
             if field not in self._values:
                 continue
             if definition['type'] in ('one2many', 'many2many'):
@@ -882,12 +912,12 @@ class Model(object):
             return
         loading = self._fields[name]['loading']
         if loading == 'eager':
-            fields = [x for x, y in self._fields.iteritems()
+            fields = [x for x, y in self._fields.items()
                     if y['loading'] == 'eager']
         if not self._fields:
             fields.append('_timestamp')
-        self._values.update(self._proxy.read([self.id], fields,
-            self._config.context)[0])
+        self._values.update(
+            self._proxy.read([self.id], fields, self._context)[0])
         for field in fields:
             if (field in self._fields
                     and self._fields[field]['type'] == 'float'
@@ -897,28 +927,27 @@ class Model(object):
 
     def _default_get(self):
         'Set default values'
-        fields = self._fields.keys()
-        self._default_set(self._proxy.default_get(fields, False,
-            self._config.context))
+        fields = list(self._fields.keys())
+        self._default_set(
+            self._proxy.default_get(fields, False, self._context))
 
     def _default_set(self, values):
         fieldnames = []
-        for field, value in values.iteritems():
+        for field, value in values.items():
             if '.' in field:
                 continue
             definition = self._fields[field]
             if definition['type'] in ('one2many', 'many2many'):
-                if value and len(value) and isinstance(value[0], (int, long)):
+                if value and len(value) and isinstance(value[0], int):
                     self._values[field] = value
                 else:
-                    relation = Model.get(definition['relation'], self._config)
-                    records = []
+                    Relation = Model.get(definition['relation'], self._config)
+                    self._values[field] = records = ModelList(
+                        definition, [], self, field)
                     for vals in (value or []):
-                        record = relation()
+                        record = Relation()
                         record._default_set(vals)
                         records.append(record)
-                    self._values[field] = ModelList(definition, records, self,
-                        field)
             else:
                 self._values[field] = value
             fieldnames.append(field)
@@ -935,7 +964,7 @@ class Model(object):
         if fields:
             definitions = ((f, self._fields[f]) for f in fields)
         else:
-            definitions = self._fields.iteritems()
+            definitions = self._fields.items()
         for field, definition in definitions:
             if not fields:
                 if field == 'id' or (skip and field in skip):
@@ -988,15 +1017,19 @@ class Model(object):
                 getattr(self, field).remove(record, _changed=False)
             if value and (value.get('add') or value.get('update')):
                 for index, vals in value.get('add', []):
-                    relation = Model.get(self._fields[field]['relation'],
-                            self._config)
-                    record = relation(_default=False)
+                    group = getattr(self, field)
+                    Relation = Model.get(
+                        self._fields[field]['relation'], self._config)
+                    config = Relation._config
+                    with config.reset_context(), \
+                            config.set_context(self._context):
+                        record = Relation(_group=group, _default=False)
                     record._set_on_change(vals)
                     # append without signal
                     if index == -1:
-                        list.append(getattr(self, field), record)
+                        list.append(group, record)
                     else:
-                        list.insert(getattr(self, field), index, record)
+                        list.insert(group, index, record)
                 for vals in value.get('update', []):
                     if 'id' not in vals:
                         continue
@@ -1004,27 +1037,31 @@ class Model(object):
                         if record.id == vals['id']:
                             record._set_on_change(vals)
         elif (self._fields[field]['type'] in ('one2many', 'many2many')
-                and len(value) and not isinstance(value[0], (int, long))):
+                and len(value) and not isinstance(value[0], int)):
             self._values[field] = []
             for vals in value:
-                relation = Model.get(self._fields[field]['relation'],
-                        self._config)
-                record = relation(_default=False, **vals)
-                getattr(self, field).append(record)
+                Relation = Model.get(
+                    self._fields[field]['relation'], self._config)
+                config = Relation._config
+                records = getattr(self, field)
+                with config.reset_context(), \
+                        config.set_context(records._get_context()):
+                    record = Relation(_default=False, **vals)
+                records.append(record)
         else:
             self._values[field] = value
         self._changed.add(field)
 
     def _set_on_change(self, values):
         later = {}
-        for field, value in values.iteritems():
+        for field, value in values.items():
             if field not in self._fields:
                 continue
             if self._fields[field]['type'] in ('one2many', 'many2many'):
                 later[field] = value
                 continue
             self._on_change_set(field, value)
-        for field, value in later.iteritems():
+        for field, value in later.items():
             self._on_change_set(field, value)
 
     def _on_change(self, names):
@@ -1038,12 +1075,12 @@ class Model(object):
             on_change = definition.get('on_change')
             if not on_change:
                 continue
-            if isinstance(on_change, basestring):
+            if isinstance(on_change, str):
                 definition['on_change'] = on_change = PYSONDecoder().decode(
                     on_change)
             values.update(self._on_change_args(on_change))
         if values:
-            context = self._config.context
+            context = self._context
             changes = getattr(self._proxy, 'on_change')(values, names, context)
             for change in changes:
                 self._set_on_change(change)
@@ -1052,7 +1089,7 @@ class Model(object):
         fieldnames = set(names)
         to_change = set()
         later = set()
-        for field, definition in self._fields.iteritems():
+        for field, definition in self._fields.items():
             on_change_with = definition.get('on_change_with')
             if not on_change_with:
                 continue
@@ -1064,14 +1101,14 @@ class Model(object):
             to_change.add(field)
             values.update(self._on_change_args(on_change_with))
         if to_change:
-            context = self._config.context
+            context = self._context
             changes = getattr(self._proxy, 'on_change_with')(values,
                 list(to_change), context)
             self._set_on_change(changes)
         for field in later:
             on_change_with = self._fields[field]['on_change_with']
             values = self._on_change_args(on_change_with)
-            context = self._config.context
+            context = self._context
             result = getattr(self._proxy, 'on_change_with_%s' % field)(values,
                     context)
             self._on_change_set(field, result)
@@ -1094,9 +1131,11 @@ class Wizard(object):
         self.form_state = None
         self.actions = []
         self._config = config or proteus.config.get_config()
-        self._context = context or {}
+        self._context = self._config.context
+        if context:
+            self._context.update(context)
         self._proxy = self._config.get_proxy(name, type='wizard')
-        result = self._proxy.create(self._config.context)
+        result = self._proxy.create(self._context)
         self.session_id, self.start_state, self.end_state = result
         self.states = [self.start_state]
         self.models = models
@@ -1109,7 +1148,6 @@ class Wizard(object):
         self.state = state
         while self.state != self.end_state:
             ctx = self._context.copy()
-            ctx.update(self._config.context)
             if self.models:
                 ctx['active_id'] = self.models[0].id
                 ctx['active_ids'] = [model.id for model in self.models]
@@ -1133,7 +1171,8 @@ class Wizard(object):
 
             if 'view' in result:
                 view = result['view']
-                self.form = Model.get(view['fields_view']['model'])()
+                self.form = Model.get(
+                    view['fields_view']['model'], self._config)()
                 self.form._default_set(view['defaults'])
                 self.states = [b['state'] for b in view['buttons']]
                 self.form_state = view['state']
@@ -1151,7 +1190,7 @@ class Wizard(object):
                 return
 
         if self.state == self.end_state:
-            self._proxy.delete(self.session_id, self._config.context)
+            self._proxy.delete(self.session_id, self._context)
             if self.models:
                 for record in self.models:
                     record.reload()
@@ -1164,7 +1203,9 @@ class Report(object):
         super(Report, self).__init__()
         self.name = name
         self._config = config or proteus.config.get_config()
-        self._context = context or {}
+        self._context = self._config.context
+        if context:
+            self._context.update(context)
         self._proxy = self._config.get_proxy(name, type='report')
 
     def execute(self, models=None, data=None):
@@ -1217,8 +1258,9 @@ def _convert_action(action, data=None, context=None, config=None):
 
         res_model = action.get('res_model', data.get('res_model'))
         res_id = action.get('res_id', data.get('res_id'))
-        Model_ = Model.get(res_model)
-        with config.set_context(action_ctx):
+        Model_ = Model.get(res_model, config)
+        config = Model_._config
+        with config.reset_context(), config.set_context(action_ctx):
             if res_id is None:
                 return Model_.find(domain)
             else:
@@ -1230,8 +1272,10 @@ def _convert_action(action, data=None, context=None, config=None):
             'context': context,
             }
         if 'model' in data:
-            Model_ = Model.get(data['model'])
-            kwargs['models'] = [Model_(id_) for id_ in data.get('ids', [])]
+            Model_ = Model.get(data['model'], config)
+            config = Model_._config
+            with config.reset_context(), config.set_context(context):
+                kwargs['models'] = [Model_(id_) for id_ in data.get('ids', [])]
         return Wizard(action['wiz_name'], **kwargs)
     elif action['type'] == 'ir.action.report':
         ActionReport = Report(action['report_name'], context=context)
